package org.jruby.webapp;
import org.apache.commons.pool.ObjectPool;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.jruby.RubyException;
import org.jruby.Ruby;
import org.jruby.runtime.Constants;
import org.jruby.javasupport.JavaEmbedUtils;
import org.jruby.exceptions.RaiseException;
import java.io.IOException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.util.NoSuchElementException;
import java.util.Collections;
import java.util.Set;
import java.util.Iterator;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletException;

/**
 * This servlet dispatches to a Ruby on Rails application, using JRuby as the interpreter.
 *
 * The dispatcher implementation is defined in subclasses.
 *
 * @author Robert Egglestone
 */
public abstract class AbstractRailsServlet extends HttpServlet {

	/** How long in seconds to suggest the client waits between requests if we're overloaded */
	private int retryAfterOverloading = 60;
	/** The pool of JRuby runtimes to process the requests */
	private ObjectPool runtimePool;

	protected abstract void dispatchRequest(Ruby runtime, HttpServletRequest request, HttpServletResponse response) throws Exception;

	/**
	 * Initialize the servlet.
	 */
	public void init() throws ServletException {
		// disable objectspace because of deadlock on tests
        if (System.getProperty("jruby.objectspace.enabled") == null) {
			System.setProperty("jruby.objectspace.enabled", "false");
        }

		// prepare JRuby home
		if (isStandalone()) {
			createJRubyHome();
		} else {
			setJRubyHome();
		}

		// find the root of the web application
		String railsRoot = getInitParameter("rails.root");
		if (railsRoot == null || railsRoot.length() == 0) {
			railsRoot = getPath("/");
		}

		// create the factory
		RailsFactory railsFactory = new RailsFactory();
		railsFactory.setRailsRoot(railsRoot);
		railsFactory.setRailsEnvironment(getDeploymentEnvironment());
		railsFactory.setServletContext(getServletContext());

		if (isStandalone()) {
			railsFactory.removeEnvironment("GEM_HOME");
			railsFactory.removeEnvironment("GEM_PATH");
		}

		initRailsFactory(railsFactory);

		// create the pool
		runtimePool = createObjectPool(railsFactory);
	}

	private void setJRubyHome() {
		String jrubyHome = getServletContext().getInitParameter("jruby.home");
		if (jrubyHome != null && jrubyHome.length() != 0) {
			System.setProperty("jruby.home", jrubyHome);
		}
	}

	private void createJRubyHome() throws ServletException {
		// create a temporary jruby home
		File jrubyHome;
		try {
			jrubyHome = createTemporaryDirectory("jruby-rails-", ".tmp", 5);
			System.setProperty("jruby.home", jrubyHome.getAbsolutePath());
		} catch (IOException e) {
			throw new ServletException("Failed to create a temporary JRuby home");
		}

		// find expected gem locations
		char s = File.separatorChar;
		File gemHome = new File(jrubyHome, "lib" + s + "ruby" + s + "gems" + s + Constants.RUBY_MAJOR_VERSION);

		// create a ruby instance for installing
		Ruby runtime = JavaEmbedUtils.initialize(Collections.EMPTY_LIST);
		runtime.evalScript("ENV['GEM_HOME'] = File.expand_path('" + rubyPath(gemHome) + "')");
		runtime.evalScript("ENV['GEM_PATH'] = ENV['GEM_HOME']");

		// extract jruby home
		if (!runtime.getLoadService().require("jruby/extract")) {
			throw new ServletException("Failed to load jruby/extract");
		}
		runtime.evalScript("JRuby::Extract.new.extract");

		// load rubygems
		if (!runtime.getLoadService().require("rubygems")) {
			throw new ServletException("Failed to load rubygems");
		}
		runtime.evalScript("Gem.manage_gems");

		runtime.getLoadService().require("fileutils");
		runtime.evalScript("unless File.respond_to?(:symlink); class File; def self.symlink(anOldName, aNewName); end; end; end");

		// install any required gems
		File gemCache = new File(gemHome, "cache");
		if (!gemCache.exists()) throw new ServletException("Could not find gemcache: " + gemCache.getAbsolutePath());
		Set gems = getServletContext().getResourcePaths("/WEB-INF/gems/");
		if (gems != null) {
			Iterator gemIt = gems.iterator();
			while (gemIt.hasNext()) {
				String localGemFile = (String)gemIt.next();
				int lastSeparator = Math.max(localGemFile.lastIndexOf('/'), localGemFile.lastIndexOf(s));
				String gemFileName = localGemFile.substring(lastSeparator + 1);
				File cacheGemFile = new File(gemCache, gemFileName);

				// install the gem if it doesn't already exist
				try {
					if (!cacheGemFile.createNewFile()) continue;
					copyGemFile(localGemFile, cacheGemFile);
				} catch (IOException e) {
					throw new ServletException("Failed to install gem for " + gemFileName, e);
				}

				// let rubygems unpack the gem
				runtime.evalScript("Gem::Installer.new('" + rubyPath(cacheGemFile) + "').install(true)");
			}
		}
	}

	private String rubyPath(File file) {
		return file.getAbsolutePath().replace('\\', '/');
	}

	public void destroy() {
		if (isStandalone()) {
			String jrubyHome = System.getProperty("jruby.home");
			delete(new File(jrubyHome));
		}
		super.destroy();
	}

	private void delete(File target) {
		if (target.isDirectory()) {
			File[] children = target.listFiles();
			for(int i=0; i<children.length; i++) {
				delete(children[i]);
			}
		}
		if (!target.delete()) {
			target.deleteOnExit();
		}
	}

	private void copyGemFile(String localGemFile, File installedGemFile) throws IOException {
		byte[] buffer = new byte[8192];
		InputStream gemIn = null;
		FileOutputStream gemOut = null;
		try {
			gemIn = getServletContext().getResourceAsStream(localGemFile);
			gemOut = new FileOutputStream(installedGemFile);
			int bytesRead;
			while ((bytesRead = gemIn.read(buffer)) != -1) {
				gemOut.write(buffer, 0, bytesRead);
			}
		} finally {
			if (gemOut != null) gemOut.close();
			if (gemIn != null) gemIn.close();
		}
	}

	private File createTemporaryDirectory(String prefix, String suffix, int attempts) throws IOException {
		if (attempts == 0) throw new IOException("Failed to create temporary directory, couldn't find an unused name!");
		String tmpdir = System.getProperty("java.io.tmpdir");
		int random = (int)(Math.random() * 10000);
		File newtmp = new File(tmpdir, prefix + random + suffix);
		if (newtmp.exists()) return createTemporaryDirectory(prefix, suffix, attempts - 1);
		newtmp.mkdirs();
		return newtmp;
	}

	protected void initRailsFactory(RailsFactory railsFactory) {
	}

	/**
	 * Create the pool of JRuby runtimes.
	 */
	protected ObjectPool createObjectPool(RailsFactory railsFactory) {
		GenericObjectPool.Config config = new GenericObjectPool.Config();
		// when the server is loaded, pausing for the request makes it worse
		// instead, we'll just fail and return a message to the user
		config.whenExhaustedAction = GenericObjectPool.WHEN_EXHAUSTED_FAIL;
		config.maxActive = getIntegerProperty("jruby.pool.maxActive", 4);
		config.maxIdle = getIntegerProperty("jruby.pool.maxIdle", 4);
		config.minIdle = getIntegerProperty("jruby.pool.minIdle", 2);
		config.timeBetweenEvictionRunsMillis = getIntegerProperty("jruby.pool.timeBetweenEvictionRunsMillis", 10000);
		// create the pool
		GenericObjectPool pool = new GenericObjectPool(railsFactory, config);
		// preload a minimum number of objects
		Runnable preloader = new Preloader(config.minIdle);
		new Thread(preloader).start();
		return pool;
	}

	/**
	 * Is the environment explicitly set to be standalone?
	 */
	public boolean isStandalone() {
		String standalone = getServletContext().getInitParameter("jruby.standalone");
		if (standalone == null) return false;
		return Boolean.parseBoolean(standalone);
	}

	protected String getDeploymentEnvironment() {
		return getServletContext().getInitParameter("rails.env");
	}

	/**
	 * Process the HTTP request.
	 */
	protected void service(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
		try {
			serviceRequest(request, response);
		} catch (ServletException e) {
			throw e;
		} catch (IOException e) {
			throw e;
		} catch (Exception e) {
			throw new ServletException(e);
		}
	}

	protected void serviceRequest(HttpServletRequest request, HttpServletResponse response) throws Exception {
		Ruby runtime = null;
		try {
			runtime = (Ruby)runtimePool.borrowObject();
			// long startTime = System.currentTimeMillis();
			dispatchRequest(runtime, request, response);
			// log("Invocation time: " + (System.currentTimeMillis() - startTime) + "ms");
		} catch (RaiseException e) {
			RubyException re = e.getException();
			re.printBacktrace(System.err);
			if (runtime == null) {
				throw new ServletException("Failed to initialise runtime", new JRubyException(e));
			} else {
				throw new ServletException("Failed to invoke rails", new JRubyException(e));
			}
		} catch (NoSuchElementException e) {
			// all objects in the pool have been used up
			log("Warning: All JRuby processes in the pool are in use");
			response.setHeader("Retry-After", Integer.toString(retryAfterOverloading));
			response.sendError(HttpServletResponse.SC_SERVICE_UNAVAILABLE, "The server is currently overloaded, please try again later");
		} finally {
			if (runtime != null) runtimePool.returnObject(runtime);
		}
	}

	/**
	 * Locate a relative webapp path on the file system.
	 */
	private String getPath(String path) throws ServletException {
		String realPath = getServletContext().getRealPath(path);
		if (realPath == null) throw new ServletException("Could not find resource " + path);
		// remove any trailing slash
		if (realPath.endsWith("/")) {
			realPath = realPath.substring(0, realPath.length() - 1);
		}
		return realPath;
	}

	/**
	 * Gets an integer variable from the context.
	 */
	protected int getIntegerProperty(String name, int defaultValue) {
		String value = getServletContext().getInitParameter(name);
		if (value == null) {
			return defaultValue;
		} else {
			return Integer.parseInt(value);
		}
	}

	/**
	 * Preload to ensure we have at least a minimum number of objects available,
	 * the evictor will create more as required if and when it starts
	 */
	private class Preloader implements Runnable {

		private int minObjects;

		public Preloader(int minObjects) {
			this.minObjects = minObjects;
		}

		public void run() {
			try {
				while (runtimePool.getNumIdle() + runtimePool.getNumActive() < minObjects) {
					runtimePool.addObject();
					// small delay between starting
					try {
						Thread.sleep(100);
					} catch (InterruptedException ignore) {
					}
				}
			} catch (Exception e) {
				log("Failed to preload JRuby: " + e.getMessage());
			}
		}
	}
}
