package org.jruby.webapp;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.channels.FileChannel;
import java.nio.ByteBuffer;
import javax.activation.FileTypeMap;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletException;
import javax.servlet.ServletContext;

/**
 * This servlet returns a static file.
 *
 * @author Robert Egglestone
 */
public class FileServlet extends HttpServlet {

	private static final String[] DEFAULT_WELCOME_FILES = { "index.jsp", "index.html", "index.htm" };

	private int bufferSize = 1024;

	private String rootPath;

	private String prefix = File.separator + "public";

	/**
	 * Initialize the servlet, and determine the webapp root.
	 */
	public void init() throws ServletException {
		// determine the root of this webapp
		ServletContext context = getServletContext();
		rootPath = context.getRealPath("/");
		if (rootPath == null) {
			throw new ServletException(
					"Cannot find the real path of this webapp, probably using a non-extracted WAR");
		}
		if (rootPath.endsWith("/")) {
			rootPath = rootPath.substring(0, rootPath.length() - 1);
		}
	}

	public String[] getWelcomeFiles() {
		String[] welcomeFiles = (String[])getServletContext().getAttribute("org.apache.catalina.WELCOME_FILES");
		if (welcomeFiles == null) welcomeFiles = DEFAULT_WELCOME_FILES;
		return welcomeFiles;
	}

	/**
	 * Transfer the file.
	 */
	protected void doGet(HttpServletRequest request,
			HttpServletResponse response) throws ServletException, IOException {
		FileChannel in = null;
		try {

			// find the location of the file
			String contextPath = request.getContextPath();
			String relativePath = request.getRequestURI().substring(
					contextPath.length());
//			log("FileServlet :: contextPath " + contextPath);
//			log("FileServlet :: relativePath " + relativePath);

			String realPath = rootPath + prefix + relativePath;
			realPath = realPath.replaceAll("\\\\", "/");
//			log("FileServlet :: realPath " + realPath);

			// check the file and open it
			File file = new File(realPath);

			// check for welcome files if it's a directory
			if (file.isDirectory()) {
				String[] welcomeFiles = getWelcomeFiles();
				for(int i=0; i<welcomeFiles.length; i++) {
					File indexFile = new File(file, welcomeFiles[i]);
					if (indexFile.isFile()) {
						file = indexFile;
						break;
					}
				}
			}

			if (!file.isFile()) {
				log("File not found: " + realPath);
				throw new FileNotFoundException(realPath);
			}

			// check for modifications
			long ifModifiedSince = request.getDateHeader("If-Modified-Since");
			long lastModified = file.lastModified();
			if (ifModifiedSince != -1 && lastModified < ifModifiedSince) {
				throw new NotModifiedException();
			}

			// setup IO streams
			ByteBuffer buffer = ByteBuffer.allocate(bufferSize);
			in = new FileInputStream(file).getChannel();

			// start returning the response
			response.setContentType(getContentTypeFor(file.getName()));
			OutputStream out = response.getOutputStream();

			// read the bytes, returning them in the response
			while (in.read(buffer) != -1) {
				out.write(buffer.array(), 0, buffer.position());
				buffer.clear();
			}
			out.close();

		} catch (NotModifiedException e) {
			response.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
		} catch (FileNotFoundException e) {
			response.sendError(HttpServletResponse.SC_NOT_FOUND);
		} catch (IOException e) {
			response.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, e
					.getMessage());
		} finally {
			try {
				if (in != null)
					in.close();
			} catch (IOException ignore) {
			}
		}
	}

	/**
	 * Static files treat GET and POST requests the same way.
	 */
	protected void doPost(HttpServletRequest request,
			HttpServletResponse response) throws ServletException, IOException {
		doGet(request, response);
	}

	/**
	 * Return the content-type the would be returned for this file name.
	 */
	public String getContentTypeFor(String fileName) {
		// quick hack for types that are necessary, but not handled
		String lowerName = fileName.toLowerCase();
		if (lowerName.endsWith(".css")) {
			return "text/css";
		} else if (lowerName.endsWith(".js")) {
			return "text/js";
		}
		// everything else
		FileTypeMap typeMap = FileTypeMap.getDefaultFileTypeMap();
		return typeMap.getContentType(fileName);
	}

	/**
	 * An exception when the source object has not been modified. While this
	 * condition is not a failure, it is a break from the normal flow of
	 * execution.
	 */
	private static class NotModifiedException extends IOException {
		public NotModifiedException() {
		}
	}

}
