package org.jruby.webapp;
import java.io.OutputStream;
import java.io.IOException;
import javax.servlet.http.HttpServletResponse;

/**
 * This is an OutputStream which parses HTTP response data, and uses it to populate a HttpServletResponse.
 */
public class HttpOutput extends OutputStream {

	private static final int STATE_CHARACTER = 0;
	private static final int STATE_FIRST_CR = 1;
	private static final int STATE_FIRST_LF = 2;
	private static final int STATE_SECOND_CR = 3;
	private static final int STATE_HEADER_END = 4;

	private StringBuffer headerBuffer;
	private HttpServletResponse response;
	private OutputStream output;
	private int state;

	public HttpOutput(HttpServletResponse response) {
		this.response = response;
		this.headerBuffer = new StringBuffer();
		this.state = STATE_CHARACTER;
	}

	private OutputStream getOutput() throws IOException {
		if (output == null) {
			writeHeaders();
			output = response.getOutputStream();
		}
		return output;
	}

	private void writeHeaders() throws IOException {
		String[] headerLines = headerBuffer.toString().split("[\r\n]+");
		for(int i=0; i<headerLines.length; i++) {
			String headerLine = headerLines[i];
			int split = headerLine.indexOf(":");
			if (split == -1) throw new IOException("Bad header line: " + headerLine);

			String header = headerLine.substring(0, split).trim();
			String value = headerLine.substring(split + 1).trim();

			if (header.toLowerCase().equals("status")) {
				response.setStatus(getStatusCode(value));
			} else if (header.toLowerCase().equals("content-type")) {
				response.setContentType(value);
			} else {
				response.addHeader(header, value);
			}
		}
	}

	private int getStatusCode(String statusValue) {
		if (statusValue.length() < 3) {
			return HttpServletResponse.SC_INTERNAL_SERVER_ERROR;
		}

		String statusCode = statusValue.substring(0, 3);
		try {
			return Integer.parseInt(statusCode);
		} catch (NumberFormatException nfe) {
			return HttpServletResponse.SC_INTERNAL_SERVER_ERROR;
		}
	}

	/**
	 * Supplied by Ashish.
	 */
	public void write(int i) throws IOException {
		if (state == STATE_HEADER_END) {
			getOutput().write(i);
			return;
		}

		// Update the state
		// State machine looks like this
		//
		//    -------->--------
		//   |      (CR)       |
		//   |                 |
		//  CR1--->---         |
		//   |        |        |
		//   ^(CR)    |(LF)    |
		//   |        |        |
		// CHAR--->--LF1--->--EOH
		//      (LF)  |  (LF)  |
		//            |(CR)    ^(LF)
		//            |        |
		//          (CR2)-->---

		if (i == '\n') {
			// LF
			switch(state) {
				case STATE_CHARACTER:
					state = STATE_FIRST_LF;
					break;
				case STATE_FIRST_CR:
					state = STATE_FIRST_LF;
					break;
				case STATE_FIRST_LF:
				case STATE_SECOND_CR:
					state = STATE_HEADER_END;
					break;
			}
		} else if (i == '\r') {
			// CR
			switch(state) {
				case STATE_CHARACTER:
					state = STATE_FIRST_CR;
					break;
				case STATE_FIRST_CR:
					state = STATE_HEADER_END;
					break;
				case STATE_FIRST_LF:
					state = STATE_SECOND_CR;
					break;
			}
		} else {
			state = STATE_CHARACTER;
		}

		// should we use a byte buffer instead?
		headerBuffer.append((char)i);
	}

	public void write(byte b[]) throws IOException {
		if (state == STATE_HEADER_END) {
			getOutput().write(b);
		} else {
			super.write(b);
		}
	}

	public void write(byte b[], int off, int len) throws IOException {
		if (state == STATE_HEADER_END) {
			getOutput().write(b, off, len);
		} else {
			super.write(b, off, len);
		}
	}

	public void flush() throws IOException {
		if (output != null) {
			output.flush();
			response.flushBuffer();
		}
	}

	public void close() throws IOException {
		getOutput().close();
		response.flushBuffer();
	}

}
