package org.jruby.webapp;
import org.apache.commons.pool.PoolableObjectFactory;
import org.jruby.RubyException;
import org.jruby.RubyArray;
import org.jruby.Ruby;
import org.jruby.exceptions.RaiseException;
import org.jruby.javasupport.JavaEmbedUtils;
import java.util.List;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.ArrayList;
import javax.servlet.ServletException;
import javax.servlet.ServletContext;
/**
 * Creates instance of the JRuby runtime and loads Rails.
 *
 * @author Robert Egglestone
 */
public class RailsFactory implements PoolableObjectFactory {

	private String railsRoot;
	private ServletContext context;
	private int objectCount;
	private String[] extraRequires;
	private Map environmentOverrides;

	public RailsFactory() {
		environmentOverrides = new HashMap();
		extraRequires = new String[0];
	}

	public void setRailsRoot(String railsRoot) {
		this.railsRoot = railsRoot;
	}

	public void setServletContext(ServletContext context) {
		this.context = context;
	}

	public void setExtraRequires(String[] extraRequires) {
		this.extraRequires = extraRequires;
	}

	public void setRailsEnvironment(String railsEnv) {
		addEnvironmentOverride("RAILS_ENV", railsEnv);
	}

	public void addEnvironmentOverride(String name, String value) {
		if (value == null) return;
		environmentOverrides.put(name, value);
	}

	public void removeEnvironment(String name) {
		environmentOverrides.put(name, "");
	}

	public void log(String message) {
		if (context == null) return;
		context.log(message);
	}

	private synchronized int nextObjectIndex() {
		return objectCount++;
	}

	/**
	 * Initialize JRuby, the Rails application, and get a reference to the dispatcher.
	 */
	public Object makeObject() throws ServletException {
		int index = nextObjectIndex();
		boolean showLoadingInfo = (index == 0);

		List loadPaths = new ArrayList();
		loadPaths.add(railsRoot);

		// start JRuby
		Ruby runtime;
		try {
			long startTime = System.currentTimeMillis();
			runtime = JavaEmbedUtils.initialize(loadPaths);
			if (showLoadingInfo) log("JRuby init time: " + (System.currentTimeMillis() - startTime) + "ms");
		} catch (RaiseException e) {
			throw new ServletException("Failed to initialize JRuby", e);
		}

		// override any environment values
		Iterator envIt = environmentOverrides.entrySet().iterator();
		while (envIt.hasNext()) {
			Map.Entry override = (Map.Entry)envIt.next();
			String value = (String)override.getValue();
			if (value != null && value.length() != 0) {
				runtime.evalScript("ENV['" + override.getKey() + "'] = '" + override.getValue() + "'");
			} else {
				runtime.evalScript("ENV.delete '" + override.getKey() + "'");
			}
		}

		// test Gems
		try {
			runtime.getLoadService().require("rubygems");
		} catch (RaiseException e) {
			throw new ServletException("Failed to load rubygems.");
		}

		// start Rails
		try {
			long startTime = System.currentTimeMillis();
			require(runtime, railsRoot + "/config/environment");
			if (showLoadingInfo) log("Rails init time: " + (System.currentTimeMillis() - startTime) + "ms");
		} catch (RaiseException e) {
			// try to put some helpful information in the logs
			RubyException re = e.getException();
			String rubyMessage = (String)JavaEmbedUtils.rubyToJava(runtime, re.message, String.class);
			String message = "Failed to load Rails: " + rubyMessage + "\n";
			RubyArray backtrace = (RubyArray)re.backtrace();
			for(Iterator traceIt = backtrace.iterator(); traceIt.hasNext(); ) {
				String traceLine = (String)traceIt.next();
				message += "\t" + traceLine + "\n";
			}
			log(message);
			// this is returned to the user
			throw new ServletException("Could not load Rails. See the logs for more details.");
		}

		// load any extra classes required
		long startTime = System.currentTimeMillis();
		for(int i=0; i<extraRequires.length; i++) {
			try {
				require(runtime, extraRequires[i]);
			} catch (RaiseException e) {
				throw new ServletException("Failed to load extra path: " + extraRequires[i], e);
			}
		}
		if (showLoadingInfo) log("Extra init time: " + (System.currentTimeMillis() - startTime) + "ms");

		log("Runtime " + index + " loaded");
		if (index == 0) log("Requests can now be processed");

		return runtime;
	}

	public boolean validateObject(Object object) {
		return true;
	}

	public void destroyObject(Object object) {
		Ruby runtime = (Ruby)object;
		JavaEmbedUtils.terminate(runtime);
	}

	public void activateObject(Object object) {
		// do nothing
	}

	public void passivateObject(Object object) {
		// do nothing
	}

	public String toUnixStyle(String path){
		return path.replaceAll("\\\\", "/");
	}

	/**
	 * Load the specified file into the JRuby runtime.
	 * Throws a ServletException if it could not be loaded.
	 */
	private void require(Ruby runtime, String file) throws ServletException {
		if (!runtime.getLoadService().require(toUnixStyle(file))) {
			throw new ServletException(file + " cannot be found");
		}
	}

}
