package org.jruby.webapp;
import org.jruby.RubyIO;
import org.jruby.RubyClass;
import org.jruby.RubyHash;
import org.jruby.RubySymbol;
import org.jruby.RubyModule;
import org.jruby.Ruby;
import org.jruby.webapp.session.SessionHolder;
import org.jruby.javasupport.JavaEmbedUtils;
import org.jruby.runtime.builtin.IRubyObject;
import org.jruby.runtime.Block;
import java.util.Enumeration;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletException;
/**
 * This is a servlet for a Ruby on Rails application, using which dispatches the request using a Java implementation.
 *
 * @author Robert Egglestone
 */
public class RailsServletJavaDispatcher extends AbstractRailsServlet {

	protected void initRailsFactory(RailsFactory railsFactory) {
		super.initRailsFactory(railsFactory);
		railsFactory.setExtraRequires(new String[] { "dispatcher" });
	}

	protected void dispatchRequest(Ruby runtime, HttpServletRequest request, HttpServletResponse response) throws Exception {
		RubyIO stdin = new RubyIO(runtime, request.getInputStream());
		RubyIO stdout = new RubyIO(runtime, new HttpOutput(response));

		runtime.defineReadonlyVariable("$stdin", stdin);
		runtime.defineReadonlyVariable("$stdout", stdout);

		// setup the environment
		setupEnvironment(runtime, request);

		// setup the default session
		setJavaSessionStoreAsDefault(runtime, request);

		// create a cgi instance to store the data
		IRubyObject cgi = createCgi(runtime);

		// get the rails dispatcher
		RubyClass dispatcher = (RubyClass)runtime.getClass("Dispatcher");
		if (dispatcher == null) throw new ServletException("The rails dispatcher could not be found");

		// dispatch the request
		Object[] args = { cgi };
		JavaEmbedUtils.invokeMethod(runtime, dispatcher, "dispatch", args, void.class);
	}

	protected IRubyObject createCgi(Ruby runtime) {
		RubyClass cgiClass = runtime.getClass("CGI");
		IRubyObject[] cgiArgs = { JavaEmbedUtils.javaToRuby(runtime, "query") };
		return cgiClass.newInstance(cgiArgs, Block.NULL_BLOCK);
	}

	protected void setupEnvironment(Ruby runtime, HttpServletRequest request) {
		RubyHash env = (RubyHash)runtime.getObject().getConstant("ENV");
		env.clear();

		String requestUri = chomp(request.getRequestURI(), "/");

		// RFC3875 The Common Gateway Interface (CGI) Version 1.1
		setEnv(env, "AUTH_TYPE", request.getAuthType());
		if (request.getContentLength() != -1) {
			setEnv(env, "CONTENT_LENGTH", String.valueOf(request.getContentLength()));
		}
		setEnv(env, "CONTENT_TYPE", request.getContentType());
		setEnv(env, "GATEWAY_INTERFACE", "CGI/1.1");
		// setEnv(env, "PATH_INFO", ?);
		setEnv(env, "PATH_TRANSLATED", requestUri);
		setEnv(env, "QUERY_STRING", request.getQueryString(), "");
		setEnv(env, "REMOTE_ADDR", request.getRemoteAddr());
		setEnv(env, "REMOTE_HOST", request.getRemoteHost());
		setEnv(env, "REMOTE_USER", request.getRemoteUser());
		setEnv(env, "REQUEST_METHOD", request.getMethod());
		// setEnv(env, "'SCRIPT_NAME'", ?);
		setEnv(env, "SERVER_ADDR", request.getServerName());
		setEnv(env, "SERVER_NAME", request.getServerName() + ":" + request.getServerPort());
		setEnv(env, "SERVER_PORT", String.valueOf(request.getServerPort()));
		setEnv(env, "SERVER_PROTOCOL", request.getProtocol());
		setEnv(env, "SERVER_SOFTWARE", getServletContext().getServerInfo());

		// Extra values that Rails can use use of
		setEnv(env, "REQUEST_URI", requestUri);
		setEnv(env, "RAILS_ENV", getDeploymentEnvironment());
		setEnv(env, "RAILS_RELATIVE_URL_ROOT", chomp(request.getContextPath(), "/"));
		if (request.getScheme().equals("https")) {
			setEnv(env, "HTTPS", "on");
		}

		// debugging
		//runtime.evalScript("require 'pp'; pp(ENV)");

		// add all the HTTP headers
		Enumeration headerNames = request.getHeaderNames();
		while (headerNames.hasMoreElements()) {
			String name = (String)headerNames.nextElement();
			String envName = "HTTP_" + name.toUpperCase().replace('-', '_');
			setEnv(env, envName, request.getHeader(name));
		}
	}

	protected void setJavaSessionStoreAsDefault(Ruby runtime, HttpServletRequest request) throws ServletException {
		RubyModule cgiRequestClass = runtime.getClassFromPath("ActionController::CgiRequest");
		if (cgiRequestClass == null) {
			log("JavaServletStore disabled because ActionController::CgiRequest could not be found");
			return;
		}
		RubyHash defaultSessionOptions = (RubyHash)cgiRequestClass.getConstant("DEFAULT_SESSION_OPTIONS");

		runtime.getLoadService().require("builtin/rails-integration/session/java_servlet_store");
		RubyModule javaServletStore = runtime.getClassFromPath("CGI::Session::JavaServletStore");
		if (javaServletStore == null) {
			log("JavaServletStore disabled because CGI::Session::JavaServletStore could not be found");
			return;
		}

		// store the default database manager
		RubySymbol databaseManagerSymbol = RubySymbol.newSymbol(runtime, "database_manager");
		defaultSessionOptions.aset(databaseManagerSymbol, javaServletStore);

		// store the java request
		RubySymbol javaRequestSymbol = RubySymbol.newSymbol(runtime, "java_request");
		IRubyObject javaRequest = JavaEmbedUtils.javaToRuby(runtime, new SessionHolder(request));
		defaultSessionOptions.aset(javaRequestSymbol, javaRequest);
	}

	protected void setEnv(RubyHash env, String name, String value, String defaultValue) {
		if (value == null) {
			value = defaultValue;
		}
		setEnv(env, name, value);
	}

	protected void setEnv(RubyHash env, String name, String value) {
		Ruby runtime = env.getRuntime();
		if (value != null) {
			IRubyObject rubyName = JavaEmbedUtils.javaToRuby(runtime, name);
			IRubyObject rubyValue = JavaEmbedUtils.javaToRuby(runtime, value);
			env.aset(rubyName, rubyValue);
		}
	}

	private String chomp(String message, String suffix) {
		if (message.endsWith(suffix)) {
			return message.substring(0, message.length() - suffix.length());
		} else {
			return message;
		}
	}

}
